from rest_framework.metadata import SimpleMetadata


class JoinBudgetMetadata(SimpleMetadata):
    def determine_metadata(self, request, view):
        metadata = super().determine_metadata(request, view)
        metadata['actions'] = {
            'POST': {
                'id': {
                    'type': 'UUID',
                    'required': True,
                    'read_only': False,
                }
            }
        }
        return metadata



# Generated by Django 4.2.4 on 2023-08-25 10:29

import budgetmanager.models
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('budgetmanager', '0002_budget_last_modified_budget_modified_by_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='ShareCode',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('can_edit', models.BooleanField(default=False)),
                ('expiry', models.DateTimeField(default=budgetmanager.models._get_sharecode_expiry, editable=False)),
                ('added_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('budget', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='budgetmanager.budget')),
            ],
        ),
    ]



'''
Model classes
'''
# pylint:disable=no-member
from datetime import datetime, timedelta
from decimal import Decimal
import uuid

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.utils import timezone


def _get_total_amount(queryset) -> Decimal:
    return queryset.filter(pending=False).aggregate(
        models.Sum('amount', default=0)
    )['amount__sum']


def _get_sharecode_expiry():
    return timezone.now() + timedelta(days=10)


class BaseModel(models.Model):
    last_modified = models.DateTimeField(auto_now=True)
    modified_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='+',
    )

    class Meta:
        abstract = True


class Budget(BaseModel):
    '''
    Model for a budget
    '''
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
    )
    name = models.CharField(max_length=100)
    description = models.TextField(null=True, blank=True)
    active = models.BooleanField(default=True)
    shared_users = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        through='BudgetShare',
        through_fields=('budget', 'user'),
        related_name='shared_budgets',
        blank=True,
    )

    def __str__(self):
        return str(self.name)

    @transaction.atomic
    def add_from_csv(self, text: str):
        '''
        Add payees and payments to this budget from a CSV formatted string
        '''
        rows = text.strip().split('\n')
        for line in rows:
            record = line.split(',')
            payee = Payee.objects.get_or_create(
                name=record[0],
                budget=self
            )[0]
            payment = Payment(
                payee=payee,
                amount=record[1],
                date=datetime.strptime(record[2], '%d/%m/%Y'),
            )
            if len(record) >= 4:
                payment.notes = record[3]
            if len(record) >= 5:
                payment.pending = record[4] != ''
            payment.save()

    @property
    def total(self):
        '''The total amount of the Payments of this Budget'''
        return _get_total_amount(Payment.objects.filter(payee__budget=self))

    @classmethod
    def get_user_model(cls):
        return cls._meta.get_field('user').related_model

    def has_access(self, user, editable: bool, requires_owner: bool = False):
        '''
        Check if user has access to this Budget
        editable: False for read access, True for write access
        '''
        res = (
            self.user == user or
            not requires_owner and self.shared_users.contains(user) and (
                (editable and self.budgetshare_set.get(
                    user=user).can_edit) or not editable
            )
        )
        return res


class BudgetShare(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL,
                             on_delete=models.CASCADE)
    budget = models.ForeignKey(Budget, on_delete=models.CASCADE)
    can_edit = models.BooleanField(default=False)
    added = models.DateTimeField(auto_now_add=True)
    added_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        related_name='+',
    )

    def clean(self):
        if self.user == self.budget.user:
            raise ValidationError('Budget owner cannot be a shared user')

    @transaction.atomic
    def transfer_budget(self):
        budget = self.budget
        old_owner = budget.user
        budget.user = self.user
        self.delete()
        budget.save()
        BudgetShare.objects.create(
            user=old_owner, budget=budget, can_edit=True)

    def __str__(self):
        return f"{self.user} {'edit' if self.can_edit else 'access'} {self.budget.name}"

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=('budget', 'user'), name='One budget membership per user per budget'),
        ]


class Payee(BaseModel):
    '''
    Model for a payee
    '''
    budget = models.ForeignKey(Budget, on_delete=models.CASCADE)
    name = models.CharField(max_length=100)
    description = models.TextField(null=True, blank=True)

    def __str__(self):
        return f'{self.name} ({self.budget.name})'

    @property
    def total(self):
        '''
        The total amount of the Payments of this Payee
        '''
        return _get_total_amount(self.payment_set)


class Payment(BaseModel):
    '''
    Model for a payment
    Requires a payee and a budget
    Has an amount and date
    '''
    payee = models.ForeignKey(
        Payee,
        on_delete=models.CASCADE,
    )
    amount = models.DecimalField(decimal_places=2, max_digits=7)
    date = models.DateField()
    pending = models.BooleanField(
        default=False, verbose_name='Exclude from total')
    notes = models.TextField(null=True, blank=True)

    @classmethod
    def get_total(cls, user):
        '''Get the total amount of the user's Payments'''
        return _get_total_amount(cls.objects.filter(payee_budget__user=user))

    def __str__(self):
        if self.amount.is_signed():
            return f'{self.payee.name}: {abs(self.amount):.2f} from {self.payee.budget.name}'
        return f'{self.payee.name}: {self.amount:.2f} to {self.payee.budget.name}'


class ShareCode(models.Model):

    id = models.UUIDField(
        primary_key=True,
        default=uuid.uuid4,
        editable=False,
    )
    budget = models.ForeignKey(
        Budget,
        on_delete=models.CASCADE,
    )
    can_edit = models.BooleanField(default=False)
    added_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        related_name='+',
    )
    expiry = models.DateTimeField(
        default=_get_sharecode_expiry,
        editable=False,
    )

    def __str__(self):
        return str(self.id)

    @transaction.atomic
    def add_user(self, user):
        BudgetShare.objects.create(
            budget=self.budget,
            user=user,
            can_edit=self.can_edit,
            added_by=self.added_by,
        )
        self.delete()



# pylint: disable=unused-argument
from rest_framework.permissions import BasePermission, SAFE_METHODS


class IsBudgetOwner(BasePermission):
    def has_object_permission(self, request, view, obj):
        return obj.has_access(
            request.user,
            request.method not in SAFE_METHODS,
            request.method == 'DELETE'
        )


class CanAccessBudgetShare(BasePermission):
    SHARED_USER_PERMISSIONS = SAFE_METHODS + ('DELETE',)

    def has_object_permission(self, request, view, obj):
        return (
            obj.user == request.user and request.method in self.SHARED_USER_PERMISSIONS
            or obj.budget.user == request.user
        )


class IsPayeeOwner(BasePermission):
    def has_object_permission(self, request, view, obj):
        return obj.has_access(request.user, request.method not in SAFE_METHODS)


class IsPaymentOwner(BasePermission):
    def has_object_permission(self, request, view, obj):
        return obj.has_access(request.user, request.method not in SAFE_METHODS)



from rest_framework.serializers import (
    HiddenField,
    ModelSerializer,
    PrimaryKeyRelatedField,
    CurrentUserDefault,
)

from . import models


class UserSerializer(ModelSerializer):
    class Meta:
        model = models.Budget.get_user_model()
        fields = (
            'id',
            'username',
            'first_name',
            'last_name',
        )


class BaseSerializer(ModelSerializer):
    modified_by = PrimaryKeyRelatedField(read_only=True)
    modified_by_hidden = HiddenField(
        default=CurrentUserDefault(),
        source='modified_by'
    )


class BudgetSerializer(BaseSerializer):
    user = PrimaryKeyRelatedField(default=CurrentUserDefault(), read_only=True)

    class Meta:
        model = models.Budget
        fields = (
            'id',
            'name',
            'description',
            'active',
            'user',
            'last_modified',
            'modified_by',
            'modified_by_hidden',
        )


class BudgetShareSerializer(ModelSerializer):
    user = PrimaryKeyRelatedField(read_only=True)
    budget = PrimaryKeyRelatedField(read_only=True)

    class Meta:
        model = models.BudgetShare
        fields = (
            'id',
            'budget',
            'user',
            'can_edit',
            'added',
            'added_by',
        )


class PayeeSerializer(BaseSerializer):
    class Meta:
        model = models.Payee
        fields = (
            'id',
            'name',
            'description',
            'budget',
            'last_modified',
            'modified_by',
            'modified_by_hidden',
        )


class PaymentSerializer(BaseSerializer):
    class Meta:
        model = models.Payment
        fields = (
            'id',
            'notes',
            'payee',
            'amount',
            'date',
            'pending',
            'last_modified',
            'modified_by',
            'modified_by_hidden',
        )


class ShareCodeSerializer(ModelSerializer):
    added_by = PrimaryKeyRelatedField(read_only=True)
    added_by_hidden = HiddenField(
        default=CurrentUserDefault(),
        source='added_by'
    )

    class Meta:
        model = models.ShareCode
        fields = (
            'id',
            'budget',
            'can_edit',
            'added_by',
            'added_by_hidden',
            'expiry',
        )



# Generated by Django 4.2.4 on 2023-08-24 16:05

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('budgetmanager', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='budget',
            name='last_modified',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name='budget',
            name='modified_by',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='budgetshare',
            name='added',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='budgetshare',
            name='added_by',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='payee',
            name='last_modified',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name='payee',
            name='modified_by',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='payment',
            name='last_modified',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name='payment',
            name='modified_by',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
    ]



# pylint: disable=no-member
import django.core.exceptions
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import mixins, status
from rest_framework.decorators import action
import rest_framework.exceptions
from rest_framework.filters import OrderingFilter, SearchFilter
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.viewsets import GenericViewSet, ModelViewSet

from . import (
    models,
    permissions,
    serializers,
)
from .metadata import JoinBudgetMetadata
from .pagination import Pagination


class TotalView(APIView):
    permission_classes = (IsAuthenticated,)

    def get(self, request):
        return Response(
            f'{models.Payment.get_total(request.user):.2f}'
        )


class PaymentRelatedMixin(ModelViewSet):
    @action(methods=('GET',), detail=True)
    def total(self, request, pk):
        return Response({
            'total': f'{self.get_object().total:.2f}'
        })


class BudgetViewSet(PaymentRelatedMixin, ModelViewSet):
    queryset = models.Budget.objects
    serializer_class = serializers.BudgetSerializer
    permission_classes = (IsAuthenticated, permissions.IsBudgetOwner)
    pagination_class = Pagination
    filter_backends = (DjangoFilterBackend, OrderingFilter, SearchFilter)
    filterset_fields = ('active',)
    ordering_fields = ('name', 'id')
    search_fields = ('name',)

    def get_queryset(self):
        return self.queryset.filter(
            Q(user=self.request.user) |
            Q(id__in=self.request.user.budgetshare_set.values('budget_id'))
        ).all()

    @action(methods=('POST',), detail=True, url_path='csv')
    def add_from_csv(self, request, pk):
        self.get_object().add_from_csv(request.data['csv'])
        return Response(None, status=status.HTTP_204_NO_CONTENT)


class BudgetShareViewSet(
    mixins.RetrieveModelMixin,
    mixins.UpdateModelMixin,
    mixins.DestroyModelMixin,
    mixins.ListModelMixin,
    GenericViewSet
):
    queryset = models.BudgetShare.objects
    serializer_class = serializers.BudgetShareSerializer
    permission_classes = (IsAuthenticated, permissions.CanAccessBudgetShare)
    pagination_class = Pagination
    filter_backends = (DjangoFilterBackend,)
    filterset_fields = ('budget', 'user')

    def get_queryset(self):
        return self.queryset.filter(
            Q(user=self.request.user) |
            Q(budget__user=self.request.user)
        ).all()

    @action(methods=('POST',), detail=True, url_path='transfer')
    def make_budget_owner(self, request, pk):
        self.get_object().transfer_budget()
        return Response(None, status=status.HTTP_204_NO_CONTENT)


class PayeeViewSet(PaymentRelatedMixin, ModelViewSet):
    queryset = models.Payee.objects
    serializer_class = serializers.PayeeSerializer
    permission_classes = (IsAuthenticated, permissions.IsPayeeOwner)
    pagination_class = Pagination
    filter_backends = (DjangoFilterBackend, OrderingFilter, SearchFilter)
    filterset_fields = ('budget',)
    ordering_fields = ('name', 'id')
    search_fields = ('name',)

    def get_queryset(self):
        return self.queryset.filter(
            Q(budget__user=self.request.user) |
            Q(budget_id__in=self.request.user.budgetshare_set.values('budget_id'))
        ).all()


class PaymentViewSet(ModelViewSet):
    queryset = models.Payment.objects
    serializer_class = serializers.PaymentSerializer
    permission_classes = (IsAuthenticated, permissions.IsPaymentOwner)
    pagination_class = Pagination
    filter_backends = (DjangoFilterBackend, OrderingFilter)
    filterset_fields = {
        'payee': ('exact',),
        'payee__budget': ('exact',),
        'pending': ('exact',),
        'amount': ('exact', 'gt', 'lt'),
        'date': ('exact', 'gt', 'lt'),
    }
    ordering_fields = ('amount', 'date', 'id')

    def get_queryset(self):
        return self.queryset.filter(
            Q(payee__budget__user=self.request.user) |
            Q(payee__budget_id__in=self.request.user.budgetshare_set.values('budget_id'))
        ).all()


class ShareCodeViewSet(
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.DestroyModelMixin,
    mixins.ListModelMixin,
    GenericViewSet
):
    queryset = models.ShareCode.objects
    serializer_class = serializers.ShareCodeSerializer
    permission_classes = (IsAuthenticated,)
    pagination_class = Pagination

    def get_queryset(self):
        return self.queryset.filter(
            Q(budget__user=self.request.user) |
            Q(budget_id__in=self.request.user.budgetshare_set.values('budget_id'))
        )


class UserViewSet(
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    GenericViewSet
):
    queryset = models.Budget.get_user_model().objects
    serializer_class = serializers.UserSerializer
    permission_classes = (IsAuthenticated,)
    pagination_class = Pagination

    def get_queryset(self):
        return self.queryset.filter(
            Q(id=self.request.user.id) |
            Q(id__in=models.BudgetShare.objects.filter(
                budget__user=self.request.user
            ).values('user_id')) |
            Q(id__in=models.BudgetShare.objects.filter(
                user=self.request.user
            ).values('budget__user_id'))
        )


class JoinBudgetView(APIView):
    permission_classes = (IsAuthenticated,)
    metadata_class = JoinBudgetMetadata

    def post(self, request):
        try:
            get_object_or_404(
                models.ShareCode,
                pk=request.data.get('id')
            ).add_user(request.user)
            return Response(None, status.HTTP_204_NO_CONTENT)
        except django.core.exceptions.ValidationError as exc:
            raise rest_framework.exceptions.ValidationError(detail={'detail':exc})
